"""NilmDB

Object that represents a NILM database file"""

# Need absolute_import so that "import nilmdb" won't pull in nilmdb.py,
# but will pull the nilmdb module instead.
from __future__ import absolute_import
import nilmdb
from nilmdb.printf import *

import tables
import time
import sys
import os
import errno

class NilmDB(object):
    def __init__(self, basepath):
        # set up path
        self.basepath = basepath.rstrip('/')

        # Create the database path if it doesn't exist
        try:
            os.makedirs(self.basepath)
        except OSError, e:
            if e.errno != errno.EEXIST:
                raise IOError("can't create tree " + self.basepath)

        # Our HD5 file goes inside it
        h5filename = os.path.abspath(self.basepath + "/data.h5")
        self.h5file = tables.openFile(h5filename, "a", "NILM Database")

        # SQLite database too
        dbfilename = os.path.abspath(self.basepath + "/data.sql")
        self.sqldb = nilmdb.NilmDBSql(dbfilename)

        self.opened = True

    def __del__(self):
        if "opened" in self.__dict__: # pragma: no cover
            fprintf(sys.stderr,
                    "error: NilmDB.close() wasn't called, path %s",
                    self.basepath)

    def close(self):
        self.sqldb.close()
        self.h5file.close()
        del self.opened

    def stream_list(self, layout = None):
        """Return list of information for all tables in the database
        that match the given layout, or all tables if layout is None.

        Format of returned data is a list of tuples:
        (path, layout)
        """
        streams = self.sqldb.stream_list(layout)
        return sorted(streams)

    def stream_create(self, path, layout_name, index = None):
        """Create a new table in the database.

        path: path to the data (e.g. '/newton/prep')

        layout_name: one of the nilmdb.layout.layouts keys, e.g. 'PrepData'

        index: layout columns listed here are marked as PyTables indices.
        If index = none, the 'timestamp' column is indexed if it exists.
        Pass an empty list to prevent indexing.
        """
        if path[0] != '/':
            raise ValueError("paths must start with /")
        [ group, node ] = path.rsplit("/", 1)
        if group == '':
            raise ValueError("Invalid path")

        # Make the group structure, one element at a time
        group_path = group.lstrip('/').split("/")
        for i in range(len(group_path)):
            parent = "/" + "/".join(group_path[0:i])
            child = group_path[i]
            try:
                self.h5file.createGroup(parent, child)
            except tables.NodeError:
                pass

        # Get description
        desc = nilmdb.layout.description(layout_name)

        # Estimated table size (for PyTables optimization purposes): assume
        # 3 months worth of data.  It's OK if this is wrong.
        exp_rows = nilmdb.layout.expected_daily_rows(layout_name) * 90

        table = self.h5file.createTable(group, node,
                                        description = desc,
                                        expectedrows = exp_rows)

        # Create indices
        try:
            if index is None and "timestamp" in table.colnames:
                index = [ "timestamp" ]
            for ind in index:
                table.cols._f_col(ind).createIndex()
        except KeyError, e:
            # Remove this table if we got an error
            self.h5file.removeNode(group, node)
            raise e

        # Insert into database once the PyTables is happy
        self.sqldb.stream_create(path, layout_name)
            
    def stream_set_metadata(self, path, data):
        """Set stream metadata from a dictionary, e.g.
           { description = 'Downstairs lighting',
             v_scaling = 123.45 }
           This replaces all existing metadata.
           """
        self.sqldb.stream_set_metadata(path, data)

    def stream_update_metadata(self, path, newdata):
        """Update stream metadata from a dictionary"""
        data = self.stream_get_metadata(path)
        data.update(newdata)
        self.sqldb.stream_set_metadata(path, data)

    def stream_get_metadata(self, path):
        """Return stream metadata as a dictionary."""
        return self.sqldb.stream_get_metadata(path)
