# -*- coding: utf-8 -*-

"""NilmDBSql

Object for managing the SQL database"""

# Need absolute_import so that "import nilmdb" won't pull in nilmdb.py,
# but will pull the nilmdb module instead.
from __future__ import absolute_import
import nilmdb
from nilmdb.printf import *

import sqlite3
import time
import sys
import os

# Don't touch old entries -- just add new ones.
schema_updates = {
    0: """
    -- All streams
    CREATE TABLE streams(
    	id INTEGER PRIMARY KEY,		-- stream ID
        path TEXT UNIQUE NOT NULL,	-- path, e.g. '/newton/prep'
        layout TEXT NOT NULL		-- one of nilmdb.layout.layouts
    );

    -- Individual timestamped ranges in those streams.
    -- For a given start_time and end_time, this tells us that the
    -- data is stored between start_pos and end_pos.
    -- Times are stored as μs since Unix epoch
    -- Positions are opaque: PyTables rows, file offsets, etc.
    CREATE TABLE ranges(
        stream_id INTEGER,
        start_time INTEGER NOT NULL,
        end_time INTEGER NOT NULL,
        start_pos INTEGER NOT NULL,
        end_pos INTEGER NOT NULL
    );
    CREATE INDEX _ranges_index ON ranges (stream_id, start_time, end_time);
    """,

    1: """
    -- Generic dictionary-type metadata that can be associated with a stream
    CREATE TABLE metadata(
    	stream_id INTEGER,
        key TEXT NOT NULL,
        value TEXT
    );
    """,
}

class NilmDBSql(object):
    verbose = 0

    def __init__(self, dbfile):
        self.dbfile = dbfile
        # We use check_same_thread = False, assuming that the rest
        # of the code (e.g. Server) will be smart and not access the
        # database from multiple threads simultaneously.
        self.con = sqlite3.connect(self.dbfile, check_same_thread = False)
        self.update_schema()

    def close(self):
        if self.con:
            self.con.commit()
            self.con.close()

    def update_schema(self):
        cur = self.con.cursor()
        version = cur.execute("PRAGMA user_version").fetchone()[0]

        while version in schema_updates:
            cur.executescript(schema_updates[version])
            version = version + 1
            if self.verbose: # pragma: no cover
                printf("Schema updated to %d\n", version)

        with self.con:
            cur.execute("PRAGMA user_version = {v:d}".format(v=version))

    def stream_list(self, requested_layout):
        """Return list of information for all tables in the database
        that match the given layout, or all tables if layout is None.

        Format of returned data is a list of tuples:
        (path, layout)
        """
        if requested_layout:
            where = "WHERE layout=?"
            params = (requested_layout,)
        else:
            where = ""
            params = ()
        result = self.con.execute("SELECT path, layout " +
                                  "FROM streams " + where, params).fetchall()
        return result        

    def stream_create(self, path, layout_name, index = None):
        """Create a new table in the database.
        path: path to the data (e.g. '/newton/prep')
        layout_name: one of the nilmdb.layout.layouts keys, e.g. 'PrepData'"""
        with self.con as con:
            con.execute("INSERT INTO streams (path, layout) VALUES (?,?)",
                        (path, layout_name))

    def stream_set_metadata(self, path, data):
        with self.con as con:
            stream_id = self.con.execute("SELECT id FROM streams WHERE path=?",
                                         (path,)).fetchone()[0]
            for key in data:
                con.execute("DELETE FROM metadata "
                            "WHERE stream_id=? AND key=?", (stream_id, key))
                con.execute("INSERT INTO metadata VALUES (?, ?, ?)",
                            (stream_id, key, data[key]))

    def stream_get_metadata(self, path):
        """Return stream metadata as a dictionary."""
        result = self.con.execute("SELECT metadata.key, metadata.value "
                                  "FROM metadata, streams "
                                  "WHERE metadata.stream_id=streams.id "
                                  "AND streams.path=?", (path,))
        data = {}
        for (key, value) in result:
            data[key] = value
        return data
